import numpy as np
import json
import plotly as py
import plotly.graph_objs as go
import urllib
from sklearn import neighbors, svm, tree, neural_network
from sklearn.model_selection import train_test_split, GridSearchCV
from sklearn.cross_validation import cross_val_score
from sklearn.metrics import accuracy_score, classification_report, cohen_kappa_score
from pprint import pprint
from random import shuffle
from sklearn.metrics import precision_score, recall_score, make_scorer
import statistics

with open("patterns.json") as patterns_file:
    patterns = json.load(patterns_file)

with open("dataset/sessions.json") as sessions_file:
    sessions = json.load(sessions_file)

with open("dataset/users.json") as users_file:
    users = json.load(users_file)

# Returns a full feature vector consisting of:
# emotion, pulse, min_pulse, max_pulse, number of misses, average dwell time,
# average flight time, average keystrokes per minute, digraphs and trigraphs
def feature_vector_as_objects(session, emotion_type):
    vector = [
              {"feature" : "user_id",
               "category" : "global",
               "value" : session["user_id"]},
              {"feature" : "experience",
               "category" : "global",
               "value" : session["experience"]},
              {"feature" : "years_of_experience",
               "category" : "global",
               "value" : session["years_of_experience"]},
              {"feature" : "years_before_university",
               "category" : "global",
               "value" : session["years_before_university"]},
              {"feature" : "first_learned",
               "category" : "global",
               "value" : session["first_learned"]},
              {"feature" : "knows_javascript",
               "category" : "global",
               "value" : session["knows_javascript"]},
              {"feature" : "gender",
               "category" : "global",
               "value" : session["gender"]},
              {"feature" : "emotion",
               "category" : "global",
               "value" : session["emotion"]},
              {"feature" : "avg_pulse",
               "category" : "global",
               "value" : normalise(45, 90, session["pulse"])},
              {"feature" : "min_pulse",
               "category" : "global",
               "value" : normalise(45, 90, session["min_pulse"])},
              {"feature" : "max_pulse",
               "category" : "global",
               "value" : normalise(45, 90, session["max_pulse"])},
              {"feature" : "number_of_misses",
               "category" : "global",
               "value" : normalise(0, 100, session["number_of_misses"])}
             ]

    if emotion_type == "bored":
        if vector[7]["value"] != "bored":
            vector[7]["value"] = "other"
    elif emotion_type == "concentrated":
        if vector[7]["value"] != "concentrated":
            vector[7]["value"] = "other"
    elif emotion_type == "confused":
        if vector[7]["value"] != "confused":
            vector[7]["value"] = "other"
    elif emotion_type == "delighted":
        if vector[7]["value"] != "delighted":
            vector[7]["value"] = "other"
    elif emotion_type == "frustrated":
        if vector[7]["value"] != "frustrated":
            vector[7]["value"] = "other"
    elif emotion_type == "surprised":
        if vector[7]["value"] != "surprised":
            vector[7]["value"] = "other"

    # Average dwell time
    dwell_time = 0
    for key in session["keylog"]:
        dwell_time += key["time_up"] - key["time_down"]
    vector.append( {
                    "feature" : "avg_dwell",
                    "category" : "global",
                    "value" : normalise(0, 150, dwell_time / len(session["keylog"]))
                   } )

    # Average flight time
    flight_time = 0
    for index, key in enumerate(session["keylog"]):
        if index < len(session["keylog"])-1:
            flight_time += session["keylog"][index+1]["time_down"] - key["time_up"]
    vector.append( {
                    "feature" : "avg_flight",
                    "category" : "global",
                    "value" : normalise(0, 150, flight_time / len(session["keylog"]))
                   } )

    # Average keystrokes per minute
    duration = session["keylog"][-1]["time_up"] - session["keylog"][0]["time_down"]
    vector.append( {
                    "feature" : "avg_keystrokes",
                    "category" : "global",
                    "value" : normalise(0, 600, len(session["keylog"]) / duration * 60000)
                   } )

    # min/max values for normalisation
    min = -100
    max = 500

    # Digraphs
    for index, key in enumerate(session["keylog"]):
        if index < len(session["keylog"])-1:
            vector.extend((
                           # first_key_down_second_key_down_time
                           {"key_codes" : [key["key_code"], session["keylog"][index+1]["key_code"]],
                            "feature" : "first_key_down_second_key_down_time",
                            "category" : "digraphs",
                            "value" : normalise(min, max, session["keylog"][index+1]["time_down"] - key["time_down"])},

                           # first_key_dwell_time
                           {"key_codes" : [key["key_code"]],
                            "feature" : "first_key_dwell_time",
                            "category" : "digraphs",
                            "value" : normalise(min, max, key["time_up"] - key["time_down"])},

                           # first_key_up_second_key_down_time
                           {"key_codes" : [key["key_code"], session["keylog"][index+1]["key_code"]],
                            "feature" : "first_key_up_second_key_down_time",
                            "category" : "digraphs",
                            "value" : normalise(min, max, session["keylog"][index+1]["time_down"] - key["time_up"])},

                            # second_key_dwell_time
                           {"key_codes" : [key["key_code"], session["keylog"][index+1]["key_code"]],
                            "feature" : "second_key_dwell_time",
                            "category" : "digraphs",
                            "value" : normalise(min, max, session["keylog"][index+1]["time_up"] - session["keylog"][index+1]["time_down"])},

                           # digraph_duration
                           {"key_codes" : [key["key_code"], session["keylog"][index+1]["key_code"]],
                            "feature" : "digraph_duration",
                            "category" : "digraphs",
                            "value" : normalise(min, max, session["keylog"][index+1]["time_up"] - key["time_down"])}
                         ))

    # Trigraphs
    for index, key in enumerate(session["keylog"]):
        if index < len(session["keylog"])-2:
            vector.extend((
                           # first_key_down_second_key_down_time
                           {"key_codes" : [key["key_code"], session["keylog"][index+1]["key_code"]],
                            "feature" : "first_key_down_second_key_down_time",
                            "category" : "trigraphs",
                            "value" : normalise(min, max, session["keylog"][index+1]["time_down"] - key["time_down"])},

                           # first_key_dwell_time
                           {"key_codes" : [key["key_code"]],
                            "feature" : "first_key_dwell_time",
                            "category" : "trigraphs",
                            "value" : normalise(min, max, key["time_up"] - key["time_down"])},

                           # first_key_up_second_key_down_time
                           {"key_codes" : [key["key_code"], session["keylog"][index+1]["key_code"]],
                            "feature" : "first_key_up_second_key_down_time",
                            "category" : "trigraphs",
                            "value" : normalise(min, max, session["keylog"][index+1]["time_down"] - key["time_up"])},

                           # second_key_down_third_key_down_time
                           {"key_codes" : [session["keylog"][index+1]["key_code"], session["keylog"][index+2]["key_code"]],
                            "feature" : "second_key_down_third_key_down_time",
                            "category" : "trigraphs",
                            "value" : normalise(min, max, session["keylog"][index+2]["time_down"] - session["keylog"][index+1]["time_down"])},

                           # second_key_dwell_time
                           {"key_codes" : [session["keylog"][index+1]["key_code"]],
                            "feature" : "second_key_dwell_time",
                            "category" : "trigraphs",
                            "value" : normalise(min, max, session["keylog"][index+1]["time_up"] - session["keylog"][index+1]["time_down"])},

                           # second_key_up_third_key_down_time
                           {"key_codes" : [session["keylog"][index+1]["key_code"], session["keylog"][index+2]["key_code"]],
                            "feature" : "second_key_up_third_key_down_time",
                            "category" : "trigraphs",
                            "value" : normalise(min, max, session["keylog"][index+2]["time_down"] - session["keylog"][index+1]["time_up"])},

                           # third_key_dwell_time
                           {"key_codes" : [session["keylog"][index+2]["key_code"]],
                            "feature" : "third_key_dwell_time",
                            "category" : "trigraphs",
                            "value" : normalise(min, max, session["keylog"][index+2]["time_up"] - session["keylog"][index+2]["time_down"])},

                           # trigraph_duration
                           {"key_codes" : [key["key_code"], session["keylog"][index+1]["key_code"], session["keylog"][index+2]["key_code"]],
                            "feature" : "trigraph_duration",
                            "category" : "trigraphs",
                            "value" : normalise(min, max, session["keylog"][index+2]["time_up"] - key["time_down"])}
                        ))

    return vector

def normalise(min, max, value):
    return value
    # return ((value - min) / (max - min))

def clense_keylog(target):
    resets = []
    target_list = []
    target_dict = []

    for i, key in enumerate(target):

        # Ignore backspace and removes the last added key
        # assuming that the backspace deleted that key
        if key["key_code"] in {8} and len(target_list) > 0:
            del target_list[-1]
            del target_dict[-1]
        # Ignore delete, backspace and space
        elif key["key_code"] in {46, 8, 32}:
            continue
        # Ignore editor resets and store the index
        elif key["key_code"] in {256}:
            resets.append(i)
        # Ignore arrow keys
        elif key["key_code"] in {32, 37, 38, 39, 40}:
            continue
        else:
            target_list.append(key["key_code"])
            target_dict.append(key)

    return resets, target_list, target_dict

# http://stackoverflow.com/questions/17870544/find-starting-and-ending-indices-of-sublist-in-list
def find_sublist(pattern, target):
    results = []
    pattern_length = len(pattern)
    for i in (j for j, e in enumerate(target) if e == pattern[0]):
        if target[i:i+pattern_length] == pattern:
            results.append((i, i+pattern_length-1))

    return results

def find_user(user_id):
    for i, user in enumerate(users):
        if user["_id"]["$oid"] == user_id:
            return user

def usable_users_set():
    user_ids = set()

    for user in users:
        if user["gender"] != "unknown":
            user_ids.add(user["_id"]["$oid"])

    print('Number of usable users:', len(user_ids))

    return user_ids

# Counts the number of instances for each pattern
def count_instances(vectors):

    number_of_instances = {}

    for i in vectors:
        print(i)
        number_of_instances[i] = {"bored" : 0,
                   "frustrated" : 0,
                   "concentrated" : 0,
                   "delighted" : 0,
                   "confused" : 0,
                   "surprised" : 0,
                   "other" : 0}

        for j in vectors[i]:
            number_of_instances[i][j[7]["value"]] += 1

    return number_of_instances

# Counts the number of instances for one pattern
def count_instances_for_pattern(vector, pulse=False):

    number_of_instances = {"bored" : 0,
               "frustrated" : 0,
               "concentrated" : 0,
               "delighted" : 0,
               "confused" : 0,
               "surprised" : 0,
               "other" : 0}

    if pulse:
        for i in vector:
            if i[8]["value"] > 0:
                number_of_instances[i[7]["value"]] += 1
    else:
        for i in vector:
            number_of_instances[i[7]["value"]] += 1


    return number_of_instances

# Creates a dictionary with all feature vectors sorted by pattern
def create_vectors(emotion_type="all"):
    usable_users = usable_users_set()
    vectors = {}

    for i, pattern in enumerate(patterns):
        vectors[pattern["text"]] = []

        for j, session in enumerate(sessions):
            # 2 keys are the shortest word (if)
            if len(session["keylog"]) > 1 and session["user_id"] in usable_users:

                cleaning = clense_keylog(session["keylog"])

                coord = find_sublist(pattern["key_codes"], cleaning[1])
                cleanlog = []

                if len(coord) > 0:

                    for j in range(coord[-1][0], coord[-1][1]+1):
                        cleanlog.append(cleaning[2][j])

                    if "keylog" in session:
                        session["keylog"] = cleanlog

                    user = find_user(session["user_id"])
                    session["gender"] = user["gender"]
                    session["experience"] = user["experience"]
                    session["years_of_experience"] = user["years_of_experience"]
                    session["years_before_university"] = user["years_before_university"]
                    session["first_learned"] = user["first_learned"]
                    session["knows_javascript"] = user["languages"]["javascript"]


                    vector = feature_vector_as_objects(session, emotion_type)

                    vectors[pattern["text"]].append(vector)

    return vectors

def classifier(vector,
               labels={"bored",
                       "concentrated",
                       "confused",
                       "delighted",
                       "frustrated",
                       "surprised",
                       "other"},
              pulse=False,
              method="knn",
              label_limit=60,
              k=0,
              user_id=""):

    # Feature vectors
    X = []

    # Labels
    y = []

    # Shuffles the vector list
    # shuffle(vector)

    status = {"bored" : 0,
              "concentrated" : 0,
              "confused" : 0,
              "delighted" : 0,
              "frustrated" : 0,
              "surprised" : 0,
              "other" : 0}
    if user_id:
        print('user_id', user_id)

    if pulse:
        for i in vector:
            # use this statement to limit users by type
            # i[5] is "knows_javascript"
            if user_id:
                if status[i[7]["value"]] < label_limit and i[0]["value"] == user_id:
                    # If there is more than label_limit instances of a label,
                    # there will not be added more isntances with that label
                    if i[8]["value"] > 0 and i[7]["value"] in labels:
                        status[i[7]["value"]] += 1
                        X.append([x["value"] for x in i[8:]])
                        y.append(i[7]["value"])

            else:
                if status[i[7]["value"]] < label_limit:
                    # If there is more than label_limit instances of a label,
                    # there will not be added more isntances with that label
                    if i[8]["value"] > 0 and i[7]["value"] in labels:
                        status[i[7]["value"]] += 1
                        X.append([x["value"] for x in i[8:]])
                        y.append(i[7]["value"])
    else:
        for i in vector:
            # use this statement to limit users by type
            if user_id:
                if i[5]["value"] > 0 and i[0]["value"] == user_id and status[i[7]["value"]] < label_limit:
                    # If there is more than label_limit instances of a label,
                    # there will not be added more isntances with that label
                    # Add "and i[8]["value"] > 0" to get vectors from those whom
                    # have measured pulse, but without the feature in the vector
                    if i[7]["value"] in labels:
                        status[i[7]["value"]] += 1
                        X.append([x["value"] for x in i[11:]])
                        y.append(i[7]["value"])
            else:
                if i[5]["value"] > 0 and status[i[7]["value"]] < label_limit:
                    # If there is more than label_limit instances of a label,
                    # there will not be added more isntances with that label
                    # Add "and i[8]["value"] > 0" to get vectors from those whom
                    # have measured pulse, but without the feature in the vector
                    if i[7]["value"] in labels:
                        status[i[7]["value"]] += 1
                        X.append([x["value"] for x in i[11:]])
                        y.append(i[7]["value"])

    print('Sessions used:', len(X))

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3, random_state=42)

    if method is "knn" and k > 0:
        # Set the parameters by cross-validation
        # http://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html
        param_grid = {'n_neighbors':(), 'weights':('uniform', 'distance'), 'p' : (1,2)}

        for i in range(3, label_limit):
            param_grid["n_neighbors"] = param_grid["n_neighbors"] + (i,)
            if i > label_limit/2 - 1 or i > 7:
                break

        print('Parameters for k:', param_grid["n_neighbors"])

        scores = ['precision', 'recall']

        for score in scores:
            print("# Tuning hyper-parameters for %s" %score)
            print()
            clf = GridSearchCV(neighbors.KNeighborsClassifier(n_neighbors=3), param_grid, cv=k, scoring="%s_macro" %score)
            clf.fit(X_train, y_train)
            print("Best parameters set found on development set:")
            print()
            print(clf.best_params_)
            print()
            print("Grid scores on development set:")
            print()
            means = clf.cv_results_['mean_test_score']
            stds = clf.cv_results_['std_test_score']
            for mean, std, params in zip(means, stds, clf.cv_results_['params']):
                print("%0.3f (+/-%0.03f) for %r"
                      % (mean, std * 2, params))
            print()

            print("Detailed classification report:")
            print()
            print("The model is trained on the full development set.")
            print("The scores are computed on the full evaluation set.")
            print()
            y_true, y_pred = y_test, clf.predict(X_test)
            print(classification_report(y_true, y_pred))
            print('Kappa score:', cohen_kappa_score(y_true, y_pred))
            print()

    elif method is "knn" and k is 0:
        # http://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html
        adapt_knn = neighbors.KNeighborsClassifier(n_neighbors=7, algorithm='auto', weights = 'distance')
        adapt_knn.fit(X_train, y_train)
        pred = adapt_knn.predict(X_test)
        print(classification_report(y_test, pred))

    elif method is "svm" and k > 0:
        # Set the parameters by cross-validation
        # http://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html
        param_grid = [{'kernel': ['rbf'], 'gamma': [1, 0.1, 0.01, 0.001, 0.0001],
                      'C': [1, 10, 30, 50]},
                      {'kernel': ['linear'], 'C': [1, 10, 30, 50]}]

        scores = ['precision', 'recall']

        for score in scores:
            print("# Tuning hyper-parameters for %s" %score)
            clf = GridSearchCV(svm.SVC(C=1), param_grid, cv=k, scoring='%s_macro' %score)
            clf.fit(X_train, y_train)
            print()
            print(clf.best_params_)
            print()
            print("Grid scores on development set:")
            print()
            means = clf.cv_results_['mean_test_score']
            stds = clf.cv_results_['std_test_score']
            for mean, std, params in zip(means, stds, clf.cv_results_['params']):
                print("%0.3f (+/-%0.03f) for %r"
                      % (mean, std * 2, params))
            print()

            print("Detailed classification report:")
            print()
            print("The model is trained on the full development set.")
            print("The scores are computed on the full evaluation set.")
            print()
            y_true, y_pred = y_test, clf.predict(X_test)
            print(classification_report(y_true, y_pred))
            print()
            print('Kappa score:', cohen_kappa_score(y_true, y_pred))
            print()

    elif method is "svm" and k is 0:
        # http://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html
        adapt_svm = svm.SVC(decision_function_shape='ovo')
        adapt_svm.fit(X_train, y_train)
        pred = adapt_svm.predict(X_test)
        print(classification_report(y_test, pred))

    elif method is "tree" and k > 0:
        # Set the parameters by cross-validation
        # http://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html
        param_grid = {'max_depth': (1, 10, 20, 40, 80, 160), 'splitter':('best', 'random')}

        scores = ['precision', 'recall']

        for score in scores:
            print("# Tuning hyper-parameters for %s" %score)
            clf = GridSearchCV(tree.DecisionTreeClassifier(), param_grid, cv=k, scoring='%s_macro' %score)
            clf.fit(X_train, y_train)
            print()
            print(clf.best_params_)
            print()
            print("Grid scores on development set:")
            print()
            means = clf.cv_results_['mean_test_score']
            stds = clf.cv_results_['std_test_score']
            for mean, std, params in zip(means, stds, clf.cv_results_['params']):
                print("%0.3f (+/-%0.03f) for %r"
                      % (mean, std * 2, params))
            print()

            print("Detailed classification report:")
            print()
            print("The model is trained on the full development set.")
            print("The scores are computed on the full evaluation set.")
            print()
            y_true, y_pred = y_test, clf.predict(X_test)
            print(classification_report(y_true, y_pred))
            print()
            print('Kappa score:', cohen_kappa_score(y_test, y_pred))
            print()

    elif method is "tree" and k is 0:
        # http://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html
        adapt_tree = tree.DecisionTreeClassifier()
        adapt_tree.fit(X_train, y_train)
        pred = adapt_tree.predict(X_test)

    else:
        return "Method doesn't exist"

    print("Instances per class used:")
    print(status)
    # http://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_recall_fscore_support.html

def vizualise_dataset(vector, labels):

    traces = []
    y_raw = range(0, len(vector[0]))
    for i in range(1, len(vector)):

        z_raw = vector[i]
        if z_raw[7]["value"] in labels:
            x = []
            y = []
            z = []
            f = []
            for j in range(11, len(z_raw)):
                z.append([z_raw[j]["value"], z_raw[j]["value"]])
                y.append([y_raw[j], y_raw[j]])
                x.append([i, i+1])
                # TODO: Show user_id and emotion also
                f.append([z_raw[j]["feature"], z_raw[j]["feature"]])
            if z_raw[7]["value"] == "bored":
                traces.append(dict(z=z, x=x, y=y,
                colorscale=[ [i, 'rgb(255,195,13)'] for i in np.arange(0,1.1,0.1) ],
                       showscale=False, type='surface', text=f, name="bored"))
            elif z_raw[7]["value"] == "concentrated":
                traces.append(dict(z=z, x=x, y=y,
                colorscale=[ [i, 'rgb(232,103,12)'] for i in np.arange(0,1.1,0.1) ],
                showscale=False, type='surface', text=f))
            elif z_raw[7]["value"] == "confused":
                traces.append(dict(z=z, x=x, y=y,
                colorscale=[ [i, 'rgb(255,0,0)'] for i in np.arange(0,1.1,0.1) ],
                showscale=False, type='surface', text=f))
            elif z_raw[7]["value"] == "delighted":
                traces.append(dict(z=z, x=x, y=y,
                colorscale=[ [i, 'rgb(182,12,232)'] for i in np.arange(0,1.1,0.1) ],
                showscale=False, type='surface', text=f))
            elif z_raw[7]["value"] == "frustrated":
                traces.append(dict(z=z, x=x, y=y,
                colorscale=[ [i, 'rgb(50,50,50)'] for i in np.arange(0,1.1,0.1) ],
                showscale=False, type='surface', text=f))
            elif z_raw[7]["value"] == "surprised":
                traces.append(dict(z=z, x=x, y=y,
                colorscale=[ [i, 'rgb(76,91,232)'] for i in np.arange(0,1.1,0.1) ],
                showscale=False, type='surface', text=f))
            else:
                traces.append(dict(z=z, x=x, y=y,
                colorscale=[ [i, 'rgb(23,255,92)'] for i in np.arange(0,1.1,0.1) ],
                showscale=False, type='surface', text=f))

    fig = { 'data':traces, 'layout':{'title':'Adapt vectors'} }
    py.offline.plot(fig, filename='adapt-vectors.html')

# DO NOT USE: Not functioning correctly
def get_user_statistics(vectors):
    vector = vectors["console.log("]
    user_statistic = {}


    for i in vector:
        if i[0]["value"] in user_statistic:
            user_statistic[i[0]["value"]][i[7]["value"]] += 1
        else:
            user_statistic[i[0]["value"]] = {"bored" : 0,
                                             "concentrated" : 0,
                                             "confused" : 0,
                                             "delighted" : 0,
                                             "frustrated" : 0,
                                             "surprised" : 0}

    return user_statistic

    # Noteworthy users
    # 58da63a30b859a68548acdaa

def remove_outliers(vectors, limit=12):
    vector = vectors["console.log("]
    print('Sessions before removing:', len(vector))
    for i, x in enumerate(vector):
        for j, y in enumerate(x):
            if j > 8:
                if y["value"] > limit:
                    del vector[i]
    print('Sessions after removing:', len(vector))
    return vector

def visualise_vectors_blotbox(vectors):
    vector_statistics = {"bored" : [],
                         "concentrated" : [],
                         "confused" : [],
                         "surprised" : [],
                         "delighted" : [],
                         "frustrated" : []}

    for t in range(0,155):
        vector_statistics["bored"].append([])
        vector_statistics["concentrated"].append([])
        vector_statistics["confused"].append([])
        vector_statistics["surprised"].append([])
        vector_statistics["delighted"].append([])
        vector_statistics["frustrated"].append([])

    for v in vectors:
        for i, feature in enumerate(v):
            if i > 10:
                vector_statistics[v[7]["value"]][i-11].append(feature["value"])

    for i, f in enumerate(vectors[0]):
        if i > 10:
            feature = ""
            keys = ""
            emotion = "surprised"
            f_mean = statistics.mean(vector_statistics[emotion][i-11])
            f_min = min(vector_statistics[emotion][i-11])
            f_max = max(vector_statistics[emotion][i-11])

            for a in f["feature"]:
                if a == "_":
                    feature += str("\\" + a)
                else:
                    feature += str(a)
            if "key_codes" in f:
                for j, key in enumerate(f["key_codes"]):
                    if j > 0:
                        keys += ", "

                    if key == 79:
                        keys += str("o")
                    elif key == 76:
                        keys += str("l")
                    elif key == 67:
                        keys += str("c")
                    elif key == 78:
                        keys += str("n")
                    elif key == 83:
                        keys += str("s")
                    elif key == 69:
                        keys += str("e")
                    elif key == 190:
                        keys += str(".")
                    elif key == 71:
                        keys += str("g")
                    elif key == 16:
                        keys += str("Shift")
                    elif key == 56:
                        keys += str("(")

            if len(keys) == 0:
                keys = ""





            print("%s & %s & %.2f & %.2f & %.2f \\\\ \\hline" % (feature, keys, f_mean, f_min, f_max))
    # for m in vector_statistics["bored"]:
    #     print("%.2f" % ( min(m) ))



def main():
    # If you want one-against-all classification, set class to be that
    # class, else set it to "all"
    emotion_type = ""

    # Arguments: vectors, threshold limit
    # Returns a set of vectors from one pattern defined in remove_outliers
    # vector = remove_outliers(create_vectors(emotion_type), 40)
    vectors = create_vectors(emotion_type)
    # vectors = create_vectors()

    # Remove labels if they are not to be classified
    labels = {"bored", "concentrated", "confused", "surprised", "delighted",
              "frustrated", "other"}

    visualise_vectors_blotbox(vectors["console.log("])

    # This is how it should look if all emotions are in labels
    # labels = {"bored", "concentrated", "confused", "surprised", "delighted",
    #           "frustrated", "other"}

    # Does not work as intended now that remove_outliers
    # only returns vectors for one pattern. Use:
    # vectors = create_vectors()
    # vector = vectors["console.log("]
    # for pattern in vectors:
    #     if len(vectors[pattern]) > 0:
    #         print(pattern, len(vectors[pattern]))
    # pprint(count_instances(vectors))
    # pprint(count_instances_for_pattern(vector))

    # Arguments: dataset, pulse, method, label_limit (number of instances per label)
    # classifier(vector, labels, False, "knn", 103, 3)
    # classifier(vector, labels, False, "svm", 242, 3)
    # classifier(vector, labels, False, "tree", 60, 5)
    # vizualise_dataset(vector, labels)
    # pprint(get_user_statistics(vector))

# Universal all against all (w/wo pulse)
def universal_all_against_all():
    vectors = remove_outliers(create_vectors(), 40)
    classifiers = ["knn", "svm"]
    label_limit = 60
    labels = {"bored", "concentrated", "confused", "surprised", "delighted", "frustrated", "other"}

    print("-- UNIVERSAL: All-against-all classification --")
    print()
    print("Number of instances per class:")
    print(count_instances_for_pattern(vectors))
    print()
    for classifier_type in classifiers:
        print(classifier_type, "with limit", label_limit)
        print("3-fold cross-validation")
        classifier(vectors, labels, False, classifier_type, label_limit, 3)

# Universal all against all with heart rate (non-heart rate sessions are excluded)
def universal_all_against_all_pulse(pulse):
    vectors = remove_outliers(create_vectors(), 40)
    labels = {"bored", "concentrated", "confused", "surprised", "delighted", "frustrated", "other"}
    classifiers = ["knn", "svm"]
    label_limit = 15

    # pulse_users = {"58f8c2580b859a4a48cb8879",
    #                "58cef4460b859a30c46fab7a",
    #                "58da63a30b859a68548acdaa",
    #                "58e4c1f20b859a646c6b6db2",
    #                "58e65a3b0b859a4a48cb8805"}

    print("-- UNIVERSAL: All-against-all pulse classification --")
    print()
    print("Number of instances per class:")
    print(count_instances_for_pattern(vectors, True))
    print('Pulse:', pulse)
    print("Classes: Bored, concentrated, delighted, frustrated")
    for classifier_type in classifiers:
        print(classifier_type, "with limit", label_limit)
        print("3-fold cross-validation")
        classifier(vectors, labels, pulse, classifier_type, label_limit, 3)

# All against all (w/wo pulse), confused and surprised are excluded
def universal_all_against_all_subset():
    vectors = remove_outliers(create_vectors(), 40)
    classifiers = ["knn", "svm"]
    labels = {"bored", "concentrated", "delighted", "frustrated"}
    label_limit = 120

    print("-- UNIVERSAL: All-against-all subset classification --")
    print("Classes: Bored, concentrated, delighted, frustrated")

    print("Number of instances per class:")
    print(count_instances_for_pattern(vectors))

    for classifier_type in classifiers:
        print(classifier_type, "with limit", label_limit)
        print("3-fold cross-validation")
        classifier(vectors, labels, False, classifier_type, label_limit, 3)

# Universal all against all with heart rate (non-heart rate sessions are excluded)
def universal_all_against_all_subset_pulse(pulse):
    vectors = remove_outliers(create_vectors(), 40)
    labels = {"bored", "concentrated", "delighted", "frustrated"}
    classifiers = ["knn", "svm"]
    label_limit = 15

    # pulse_users = {"58f8c2580b859a4a48cb8879",
    #                "58cef4460b859a30c46fab7a",
    #                "58da63a30b859a68548acdaa",
    #                "58e4c1f20b859a646c6b6db2",
    #                "58e65a3b0b859a4a48cb8805"}

    print("-- UNIVERSAL: All-against-all subset with pulse classification --")
    print()
    print("Number of instances per class:")
    print(count_instances_for_pattern(vectors, True))
    print("Pulse:", pulse)
    print("Classes: Bored, concentrated, delighted, frustrated")
    for classifier_type in classifiers:
        print(classifier_type, "with limit", label_limit)
        print("3-fold cross-validation")
        classifier(vectors, labels, pulse, classifier_type, label_limit, 3)

# Universal one against all
def universal_one_against_all():
    emotion_types = ["bored", "confused", "delighted", "concentrated", "frustrated", "surprised"]
    limits = [60, 22, 106, 242, 84, 7]
    classifiers = ["knn", "svm"]

    labels = {"bored", "concentrated", "confused", "surprised", "delighted", "frustrated", "other"}

    print("-- UNIVERSAL: One-against-all classification --")

    for classifier_type in classifiers:
        for i, emotion in enumerate(emotion_types):
            vectors = remove_outliers(create_vectors(emotion), 40)
            print()
            print("Number of instances per class:")
            print(count_instances_for_pattern(vectors))
            print()
            print('#########', emotion)
            print(classifier_type, "with limit", limits[i])
            print("3-fold cross-validation")
            classifier(vectors, labels, False, classifier_type, limits[i], 3)

# Universal one against all with heart rate (non-heart rate sessions are excluded)
def universal_one_against_all_pulse(pulse):
    # There are 0 instances of surprised and only 4 of confused with pulse
    # surprised and confused are excluded
    emotion_types = ["bored", "delighted", "concentrated", "frustrated"]
    label_limits = [12, 11, 60, 14]
    classifiers = ["knn", "svm"]

    # pulse_users = {"58f8c2580b859a4a48cb8879",
    #                "58cef4460b859a30c46fab7a",
    #                "58da63a30b859a68548acdaa",
    #                "58e4c1f20b859a646c6b6db2",
    #                "58e65a3b0b859a4a48cb8805"}

    labels = {"bored", "concentrated", "confused", "delighted", "frustrated", "other"}

    print("-- UNIVERSAL: One-against-all with pulse classification --")
    print("Pulse:", pulse)
    for classifier_type in classifiers:
        for i, emotion in enumerate(emotion_types):
            vectors = remove_outliers(create_vectors(emotion), 40)
            print()
            print("Number of instances per class:")
            print(count_instances_for_pattern(vectors, True))
            print()
            print('#########', emotion)
            print(classifier_type, "with limit", label_limits[i])
            print("3-fold cross-validation")
            classifier(vectors, labels, pulse, classifier_type, label_limits[i], 3)


def individual_all_against_all():
    # Users with many instances + those with pulse
    individuals = ["58e504910b859a4a48cb877b", #  92 instances
                   "590093ca0b859a4a48cb88bb", #  75 instances
                   "58cef42e0b859a30c46fab79", # 100 instances
                   "58f8c2580b859a4a48cb8879", #  62 instances
                   "58cef4460b859a30c46fab7a", #  60 instances
                   "58da63a30b859a68548acdaa", #  86 instances
                   "58e4c1f20b859a646c6b6db2", #  42 instances
                   "58e65a3b0b859a4a48cb8805"] #  61 instances
    label_limits = [10, 10, 15, 30, 10, 20, 10, 10]
    classifiers = ["knn", "svm"]
    labels = {"bored", "concentrated", "confused", "surprised", "delighted", "frustrated", "other"}
    vectors = remove_outliers(create_vectors(), 40)

    print("-- INDIVIDUAL: All-against-all classification --")
    for i, individual in enumerate(individuals):
        for classifier_type in classifiers:
            classifier(vectors, labels, False, classifier_type, label_limits[i], 3, individual)

def individual_all_against_all_pulse(pulse):
    # Users with pulse
    individuals = ["58f8c2580b859a4a48cb8879", #  62 instances
                   "58cef4460b859a30c46fab7a", #  60 instances
                   "58da63a30b859a68548acdaa", #  86 instances
                   "58e4c1f20b859a646c6b6db2", #  42 instances
                   "58e65a3b0b859a4a48cb8805"] #  61 instances
    label_limits = [10, 10, 15, 30, 10, 20, 10, 10]
    classifiers = ["knn", "svm"]
    labels = {"bored", "concentrated", "confused", "surprised", "delighted", "frustrated", "other"}
    vectors = remove_outliers(create_vectors(), 40)

    print("-- INDIVIDUAL: All-against-all with pulse classification --")
    print("Pulse:", pulse)
    for i, individual in enumerate(individuals):
        for classifier_type in classifiers:
            classifier(vectors, labels, pulse, classifier_type, label_limits[i], 3, individual)

def individual_one_against_all():
    # Users with many instances + those with pulse
    individuals = ["58e504910b859a4a48cb877b", #  92 instances
                   "590093ca0b859a4a48cb88bb", #  75 instances
                   "58cef42e0b859a30c46fab79", # 100 instances
                   "58f8c2580b859a4a48cb8879", #  62 instances
                   "58cef4460b859a30c46fab7a", #  60 instances
                   "58da63a30b859a68548acdaa", #  86 instances
                   "58e4c1f20b859a646c6b6db2", #  42 instances
                   "58e65a3b0b859a4a48cb8805"] #  61 instances
    emotion_types = ["bored", "confused", "delighted", "concentrated", "frustrated", "surprised"]
    label_limits = [[4, 9, 4, 21, 52, 2],
                    [1, 3, 4, 39, 28, 0],
                    [0, 15, 63, 7, 15, 0],
                    [0, 0, 0, 37, 25, 0],
                    [3, 1, 2, 45, 9, 0],
                    [4, 5, 18, 39, 19, 1],
                    [2, 4, 1, 33, 1, 1],
                    [5, 5, 3, 34, 14, 0]]
    classifiers = ["knn", "svm"]
    labels = {"bored", "concentrated", "confused", "surprised", "delighted", "frustrated", "other"}

    print("-- INDIVIDUAL: One-against-all classification --")
    for i, individual in enumerate(individuals):
        for classifier_type in classifiers:
            for j, emotion in enumerate(emotion_types):
                if label_limits[i][j] > 9:
                    vectors = remove_outliers(create_vectors(emotion), 40)
                    print()
                    print("Number of instances per class:")
                    print(count_instances_for_pattern(vectors, True))
                    print()
                    print('#########', emotion)
                    print(classifier_type, "with limit", label_limits[i][j])
                    print("3-fold cross-validation")
                    classifier(vectors, labels, False, classifier_type, label_limits[i][j], 3, individual)

def individual_one_against_all_pulse(pulse):
    # Users with pulse
    individuals = ["58f8c2580b859a4a48cb8879", #  62 instances
                   "58cef4460b859a30c46fab7a", #  60 instances
                   "58da63a30b859a68548acdaa", #  86 instances
                   "58e4c1f20b859a646c6b6db2", #  42 instances
                   "58e65a3b0b859a4a48cb8805"] #  61 instances
    emotion_types = ["bored", "confused", "delighted", "concentrated", "frustrated", "surprised"]
    label_limits = [[0, 0, 0, 37, 25, 0],
                    [3, 1, 2, 45, 9, 0],
                    [4, 5, 18, 39, 19, 1],
                    [2, 4, 1, 33, 1, 1],
                    [5, 5, 3, 34, 14, 0]]
    classifiers = ["knn", "svm"]
    labels = {"bored", "concentrated", "confused", "surprised", "delighted", "frustrated", "other"}

    print("-- INDIVIDUAL: One-against-all with pulse classification --")
    print("Pulse:", pulse)
    for i, individual in enumerate(individuals):
        for classifier_type in classifiers:
            for j, emotion in enumerate(emotion_types):
                if classifier_type == "knn" and label_limits[i][j] > 5 or classifier_type == "svm" and label_limits[i][j] > 19:
                    vectors = remove_outliers(create_vectors(emotion), 40)
                    print()
                    print("Number of instances per class:")
                    print(count_instances_for_pattern(vectors, True))
                    print()
                    print('#########', emotion)
                    print(classifier_type, "with limit", label_limits[i][j])
                    print("3-fold cross-validation")
                    classifier(vectors, labels, pulse, classifier_type, label_limits[i][j], 3, individual)


main()
# universal_all_against_all()
# universal_all_against_all_pulse(True)
# universal_all_against_all_pulse(False)
# universal_all_against_all_subset()
# universal_all_against_all_subset_pulse(True)
# universal_all_against_all_subset_pulse(False) # Remember to add "and i[8]["value"] > 0" on lines 409 and 419
# universal_one_against_all()
# universal_one_against_all_pulse(True)
# universal_one_against_all_pulse(False) # Remember to add "and i[8]["value"] > 0" on lines 409 and 419
# individual_all_against_all()
# individual_all_against_all_pulse(True)
# individual_all_against_all_pulse(False) # Remember to add "and i[8]["value"] > 0" on lines 409 and 419
# individual_one_against_all()
# individual_one_against_all_pulse(True)
# individual_one_against_all_pulse(False)
